import { SelectQueryBuilder } from 'typeorm/query-builder/SelectQueryBuilder';

export enum JoinType {
    LeftJoinAndSelect,
    LeftJoin,
    InnerJoin,
    InnerJoinAndSelect
}

export class QueryFilters {
    constructor(joinFilters: JoinFilter[], havingFilters: HavingFilter[]) {
        this.joinFilters = joinFilters;
        this.havingFilters = havingFilters;
    }
    joinFilters: JoinFilter[];
    havingFilters: HavingFilter[];
}

export class OrderByAction {
    constructor(property: string, type: 'ASC' | 'DESC') {
        this.property = property;
        this.type = type;
    }
    property: string;
    type: 'ASC' | 'DESC';
}

export class LimitOffsetAction {
    constructor(limit?: number, offset?: number) {
        this.limit = limit;
        this.offset = offset;
    }
    limit?: number;
    offset?: number;
}

export class Parameter {
    constructor(key: string, value: any) {
        this.key = key;
        this.value = value;
    }
    key: string;
    value: string;
}

export class JoinFilter {
    constructor(property: string, alias: string, condition?: string, parameters?: Parameter[],joinType?: JoinType) {
        this.property = property;
        this.alias = alias;
        this.condition = condition;
        this.parameters = parameters;
        this.joinType = joinType;
    }
    property: string;
    alias: string;
    condition?: string;
    parameters?: Parameter[];
    joinType?: JoinType;
}

export class HavingFilter {
    constructor(having: string, parameters?: Parameter[], inclusive?: boolean, select?: boolean) {
        this.having = having;
        this.parameters = parameters;
        this.inclusive = inclusive ? inclusive : false;
        this.select = select ? select : false;
    }
    having: string;
    parameters?: Parameter[];
    inclusive: boolean;
    select: boolean;
}

export class QueryUtils {
    public static defineFilters(
        qb: SelectQueryBuilder<any>,
        joinFilters?: JoinFilter[],
        orderByActions?: OrderByAction[],
        limitOffsetAction?: LimitOffsetAction,
        havingFilters?: HavingFilter[]) {
        if (joinFilters) {
            for (const filter of joinFilters) {
                switch (filter.joinType) {
                    case JoinType.InnerJoin:
                        qb.innerJoin(filter.property, filter.alias, filter.condition);
                        break;
                    case JoinType.InnerJoinAndSelect:
                        qb.innerJoinAndSelect(filter.property, filter.alias, filter.condition);
                        break;
                    case JoinType.LeftJoin:
                        qb.leftJoin(filter.property, filter.alias, filter.condition);
                        break;
                    case JoinType.LeftJoinAndSelect:
                        qb.leftJoinAndSelect(filter.property, filter.alias, filter.condition);
                        break;
                    default:
                        break;
                }
                if (filter.parameters) {
                    for (const parameter of filter.parameters) {
                        qb.setParameter(parameter.key, parameter.value);
                    }
                }
            }
        }
        if (orderByActions) {
            for (const orderByAction of orderByActions) {
                qb.orderBy(orderByAction.property, orderByAction.type);
            }
        }
        if (limitOffsetAction) {
            if (limitOffsetAction.limit) {
                if (joinFilters) {
                    qb.take(limitOffsetAction.limit);
                } else {
                    qb.limit(limitOffsetAction.limit);
                }
            }
            if (limitOffsetAction.offset) {
                if (joinFilters) {
                    qb.skip(limitOffsetAction.offset - 1);
                } else {
                    qb.offset(limitOffsetAction.offset);
                }
            }
        }
        let first = true;
        if (havingFilters) {
            for (const have of havingFilters) {
                if (first) {
                    qb.where(have.having);
                    first = false;
                } else if (have.inclusive) {
                    qb.andWhere(have.having);
                } else {
                    qb.orWhere(have.having);
                }
                if (have.parameters) {
                    for (const parameter of have.parameters) {
                        qb.setParameter(parameter.key, parameter.value);
                    }
                }
            }
        }
    }
}
